/*-
 * Copyright (c) 2008 Vincenzo Iozzo <snagg@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include "audit_pipe_regression_test_utils.h"

struct setaudit_record
{
	int auid;
	unsigned int mask_s;
	unsigned int mask_f;
	int port;
	int machine;
	int asid;
	char err_val[256];
	int ret;
	int index;
	struct setaudit_record *next;
};

struct setaudit_record *
add_field(struct setaudit_record *head, struct setaudit_record *new)
{
	struct setaudit_record *tmp;

	tmp = head;
	if(tmp == NULL) {
		new->next = NULL;
		return (new);
	}

	/* Reach the last element of the list*/
	for(; tmp->next != NULL;tmp = tmp->next);

	new->index = tmp->index +1;
	tmp->next= new;
	new->next = NULL;

	return (head);
}

void test_setaudit()
{
	int ret;
  	auditinfo_t au;
	
	init_channel_primary();
	
	au.ai_auid = 1;
	au.ai_mask.am_success =  au.ai_mask.am_failure = 0x00001100;
	au.ai_termid.port = au.ai_termid.machine = 0;
	au.ai_asid = 0;
	
	ret = setaudit(&au);
	
  	write_int(au.ai_auid, "auid");
  	write_int(au.ai_mask.am_success, "mask_s");
	write_int(au.ai_mask.am_failure, "mask_f");
	write_int(au.ai_termid.port, "port");
	write_int(au.ai_termid.machine, "machine");
	write_int(au.ai_asid, "asid");
	write_int(ret, "ret");
	if (ret == -1)
		write_string(strerror(errno), "ret_val");
	else
		write_string("success", "ret_val");
	
	write_end();
	
	setuid(1001);
	ret = setaudit(&au);
	
  	write_int(au.ai_auid, "auid");
  	write_int(au.ai_mask.am_success, "mask_s");
	write_int(au.ai_mask.am_failure, "mask_f");
	write_int(au.ai_termid.port, "port");
	write_int(au.ai_termid.machine, "machine");
	write_int(au.ai_asid, "asid");
	write_int(ret, "ret");
	if (ret == -1)
		write_string(strerror(errno), "ret_val");
	else
		write_string("success", "ret_val");
	
	write_end();
	
	end_channel();
	
	/* Let auditpipe flush all data */
	sleep(5);
	return;
}

void validate(FILE *f, struct setaudit_record *head, char *path, pid_t pid)
{
	int i, ret, reclen, bytesread, to_parse, arg_counter;
	struct setaudit_record *tmp, *elem;
	int err = -1;
	tokenstr_t tok;
	u_char *buf;
	u_char type = 0;
	FILE *fp;
	long control_flag;
	
	to_parse = WAITING;
	arg_counter = 0;
	control_flag = 0;
	
	fp = fopen(path, "r");
	if(fp == NULL)
		return;

	tmp = head;
	if(tmp == NULL)
		return;

	while(1) {

		to_parse = WAITING;
		if(!err || reclen == -1)
			break;

		/* Record must begin with a header token. */
		do {
			type = fgetc(fp);
		} while(type != AU_HEADER_32_TOKEN);
		ungetc(type, fp);

		while ((reclen = au_read_rec(fp, &buf)) != -1) {
			bytesread = 0;
			while (bytesread < reclen) {

				/* Is this an incomplete record? */
				if (au_fetch_tok(&tok, buf + bytesread,
					reclen - bytesread) == -1) {
					err = 0;
					break;
					}	
				bytesread += tok.len;

				/* Check only setaudit syscalls */
				if(tok.tt.hdr32.e_type == 133)
					to_parse = VALID;
					
				if(to_parse == VALID) {
					switch(tok.id) {
					
					case AUT_SUBJECT32:
						TOKEN_FLAG_SET(control_flag, TOKEN_SUBJECT);
						ret = check_priv(tok.tt.proc32_ex, pid);
						if(!ret)
							report_error(tok, f);
						break;

					case AUT_ARG32:
						TOKEN_FLAG_SET(control_flag, TOKEN_ARG);
						arg_counter++;
						switch(arg_counter) {
							case 1:
								ret = check_arg(tok.tt.arg32, tmp->auid);
								break;
							
							case 2:
								ret = check_arg(tok.tt.arg32, tmp->port);
								break;
							
							case 3:
								ret = check_arg(tok.tt.arg32, tmp->machine);
								break;
							
							case 4:
								ret = check_arg(tok.tt.arg32, tmp->mask_s);
								break;
							
							case 5:
								ret = check_arg(tok.tt.arg32, tmp->mask_f);
								break;
								
							case 6:
								ret = check_arg(tok.tt.arg32, tmp->asid);
								break;
						}	
						if(!ret)
							report_error(tok, f);
						break;

					case AUT_RETURN32:
						TOKEN_FLAG_SET(control_flag, TOKEN_RETURN);
						ret = check_ret(tok.tt.ret32, tmp->ret, tmp->err_val);
						if(!ret)
							report_error(tok, f);
						break;

					case AUT_TRAILER:
						to_parse = WAITING;
						if (!TOKEN_FLAG_ISSET(control_flag, TOKEN_SUBJECT))
							report_error_string("Missing Subject token", f);
						if (!TOKEN_FLAG_ISSET(control_flag, TOKEN_ARG) ||
							arg_counter != 6)
							report_error_string("Missing Argument token", f);
						if(!TOKEN_FLAG_ISSET(control_flag, TOKEN_RETURN))
							report_error_string("Missing Return token", f);
						
						control_flag = 0;
						arg_counter = 0;
						if(tmp->next != NULL)
							tmp = tmp->next;
						break;

					default:
						break;
					}		
				}
			}
		}
	}

free(buf);
fclose(fp);
}

int
main(int argc, char *argv[])
{
	int i, quit, fd, value, index, fdout, count;
	char *shared_string, *descr, path[512];
	pid_t pid;
	struct setaudit_record *head, *elem;
	FILE *f;
	
	index = 0;
	head = NULL;
	
	init_channel();

	pid = fork();
	if(!pid) {
		sleep(15);
		test_setaudit();
	} else {
		fd = setup_auditpipe(pid);
		snprintf(path, 512, "/tmp/audit-%d", pid);

		fdout = open(path, O_RDWR | O_CREAT);
		if(fdout == -1)
			err(-1, "setaudit");

		quit = 0;
		while (!waitpid(pid, &quit, WNOHANG)) {
			
			/* Audit pipe input. */
			read_auditpipe(fd, fdout);
				
			/* 
			 * See whether is there anything on the shared-memory,
			 * if so build a structure 
			 */
			sem_getvalue(mutex, &count);
			if(count) {
				shared_string = read_string();
				elem = malloc(sizeof(struct setaudit_record));
				if(elem == NULL)
						err(-1, "MALLOC");

				while((count = parse_string(shared_string)) != -1) {
					descr = get_descr(shared_string);
					if(count == INT_TYPE) {
						if(!strncmp(descr, "auid", 4))
							elem->auid = get_int(shared_string);
						else if (!strncmp(descr, "mask_s", 6))
							elem->mask_s = get_int(shared_string);
						else if (!strncmp(descr, "mask_f", 6))
							elem->mask_f = get_int(shared_string);
						else if(!strncmp(descr, "machine", 7))
							elem->machine = get_int(shared_string);
						else if(!strncmp(descr, "port", 4))
							elem->port = get_int(shared_string);
						else if(!strncmp(descr, "asid", 4))
							elem->asid = get_int(shared_string);
						else
							elem->ret = get_int(shared_string);
					}		
					else if ( count == STRING_TYPE) 
						snprintf(elem->err_val, sizeof(elem->err_val) > 
								    strlen(shared_string) +1 ? 
								    strlen(shared_string) +1 : 
								    sizeof(elem->err_val), "%s", 
									shared_string);
					
					shared_string = read_string();
				}		
				head = add_field(head, elem);
				if(head == NULL)
					err(-1, "List error");
			}
		}
	}
	close_auditpipe(fd, fdout);
	end_channel();
	f = init_log(pid);
	validate(f, head, path, pid);
	fclose(f);
	return (0);
}
