/*-
 * Copyright (c) 2008 Vincenzo Iozzo <snagg@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include "audit_pipe_regression_test_utils.h"

struct execve_record
{
	char path[256];
	char arg[2][256];
	char err_val[256];
	int ret;
	int index;
	struct execve_record *next;
};

struct execve_record *
add_field(struct execve_record *head, struct execve_record *new)
{
	struct execve_record *tmp;

	tmp = head;
	if(tmp == NULL) {
		new->next = NULL;
		return (new);
	}

	/* Reach the last element of the list*/
	for(; tmp->next != NULL;tmp = tmp->next);

	new->index = tmp->index +1;
	tmp->next= new;
	new->next = NULL;

	return (head);
}

void test_execve()
{
	int ret;
	char **arg;
	char path[256];
	
	arg = malloc(sizeof(char));
	if (arg == NULL)
		err(-1, "MALLOC");
	snprintf(path, 256, "/sbin/ping");
	arg[0] = path;
	arg[1] = "127.0.0.1";
	arg[2] = NULL;
	
	init_channel_primary();

	write_string(arg[0], "arg0");
	write_string(arg[1], "arg1");
	write_string(path, "path");
	write_int(0, "ret");
	write_string("success", "ret_val");
	
	write_end();
	
	end_channel();
	
	/* Let auditpipe flush all data */
	sleep(5);
	
	ret = execve(path, arg, NULL);
	
	return;
}

void validate(FILE *f, struct execve_record *head, char *path, pid_t pid)
{
	int i, ret, reclen, bytesread, to_parse;
	struct execve_record *tmp, *elem;
	int err = -1;
	tokenstr_t tok;
	u_char *buf;
	u_char type = 0;
	FILE *fp;
	long control_flag;
	
	control_flag = 0;
	to_parse = WAITING;
	
	fp = fopen(path, "r");
	if(fp == NULL)
		return;

	tmp = head;
	if(tmp == NULL)
		return;

	while(1) {

		to_parse = WAITING;
		if(!err || reclen == -1)
			break;

		/* Record must begin with a header token. */
		do {
			type = fgetc(fp);
		} while(type != AU_HEADER_32_TOKEN);
		ungetc(type, fp);

		while ((reclen = au_read_rec(fp, &buf)) != -1) {
			bytesread = 0;
			while (bytesread < reclen) {

				/* Is this an incomplete record? */
				if (au_fetch_tok(&tok, buf + bytesread,
					reclen - bytesread) == -1) {
					err = 0;
					break;
					}	
				bytesread += tok.len;

				/* Check only execve syscalls. */
				if(tok.tt.hdr32.e_type == 23)
					to_parse = VALID;
					
				if(to_parse == VALID) {
					switch(tok.id) {
					case AUT_SUBJECT32:
						TOKEN_FLAG_SET(control_flag, TOKEN_SUBJECT);
						ret = check_priv(tok.tt.proc32_ex, pid);
						if(!ret)
							report_error(tok, f);
						break;
					
					case AUT_PATH:
						TOKEN_FLAG_SET(control_flag, TOKEN_PATH);
						ret = check_path(tok.tt.path, tmp->path);
						if(!ret)
							report_error(tok, f);
						break;
					
					case AUT_EXEC_ARGS:
						TOKEN_FLAG_SET(control_flag, TOKEN_EXEC_ARG);
						for( i = 0; i < tok.tt.execarg.count; i++)
							if(!(ret = check_exec_arg(tok.tt.execarg.text[i], tmp->arg[i])))
								break;
							
						if(!ret)
							report_error(tok, f);
						break;
							
					case AUT_RETURN32:
						TOKEN_FLAG_SET(control_flag, TOKEN_RETURN);
						ret = check_ret(tok.tt.ret32, tmp->ret, tmp->err_val);
						if(!ret)
							report_error(tok, f);
						break;

					case AUT_TRAILER:
						to_parse = WAITING;
						if (!TOKEN_FLAG_ISSET(control_flag, TOKEN_SUBJECT))
							report_error_string("Missing Subject token", f);
						if (!TOKEN_FLAG_ISSET(control_flag, TOKEN_EXEC_ARG))
							report_error_string("Missing Execve Argument token", f);
						if(!TOKEN_FLAG_ISSET(control_flag, TOKEN_RETURN))
							report_error_string("Missing Return token", f);
						if (!TOKEN_FLAG_ISSET(control_flag, TOKEN_PATH))
							report_error_string("Missing Path token", f);
						
						control_flag = 0;
						if(tmp->next != NULL)
							tmp = tmp->next;
						break;

					default:
						break;
					}		
				}
			}
		}
	}

free(buf);
fclose(fp);
}

int
main(int argc, char *argv[])
{
	int i, quit, fd, value, index, fdout, count;
	char *shared_string, *descr, path[512];
	pid_t pid;
	struct execve_record *head, *elem;
	FILE *f;
	
	index = 0;
	head = NULL;
	
	init_channel();

	pid = fork();
	
	if(!pid) {
		sleep(15);
		test_execve();
	} else {
		fd = setup_auditpipe(pid);
		snprintf(path, 512, "/tmp/audit-%d", pid);

		fdout = open(path, O_RDWR | O_CREAT);
		if(fdout == -1)
			err(-1, "OPEN");

		quit = 0;
		while (!waitpid(pid, &quit, WNOHANG)) {
			
			/* Audit pipe input. */
			read_auditpipe(fd, fdout);
				
			/* 
			 * See whether is there anything on the shared-memory,
			 * if so build a structure 
			 */
			sem_getvalue(mutex, &count);
			if(count) {
				shared_string = read_string();
				elem = malloc(sizeof(struct execve_record));
				if(elem == NULL)
						err(-1, "MALLOC");

				while((count = parse_string(shared_string)) != -1) {
					descr = get_descr(shared_string);
					if(count == INT_TYPE)
						elem->ret = get_int(shared_string);
					else if ( count == STRING_TYPE) {
						if(!strncmp(descr, "path", 4))
							snprintf(elem->path, sizeof(elem->path) > 
								    strlen(shared_string) +1 ? 
								    strlen(shared_string) +1 
								    : sizeof(elem->path), "%s", 
									shared_string);
						else if(!strncmp(descr, "arg0", 4))
							snprintf(elem->arg[0], sizeof(elem->arg[0]) > 
								    strlen(shared_string) +1 ? 
								    strlen(shared_string) +1 
								    : sizeof(elem->arg[0]), "%s", 
									shared_string);
						else if(!strncmp(descr, "arg1", 4))
							snprintf(elem->arg[1], sizeof(elem->arg[1]) > 
								    strlen(shared_string) +1 ? 
								    strlen(shared_string) +1 
								    : sizeof(elem->arg[1]), "%s", 
									shared_string);			
						else
							snprintf(elem->err_val, sizeof(elem->err_val) > 
									strlen(shared_string) +1 ? 
									strlen(shared_string) +1 : 
									sizeof(elem->err_val), "%s", 
									shared_string);
					}
					shared_string = read_string();
				}		
				head = add_field(head, elem);
				if(head == NULL)
					err(-1, "List error");
			}
		}
		/* One last read to catch execve*/
		read_auditpipe(fd, fdout);
	}
	close_auditpipe(fd, fdout);
	end_channel();
	f = init_log(pid);
	validate(f, head, path, pid);
	fclose(f);
	return (0);
}
